import unittest

import numpy as np
from cbp.graph import GOGraph
from cbp.node import GOFactorNode, GOVarNode, VarNode, FactorNode
from scipy.stats import norm


class TestGOGraph(unittest.TestCase):
    def setUp(self):
        self.rng = np.random.RandomState(1)

    def test_factor_discrete(self):
        num_obser = 5
        dim = 10
        connected_names = ['VarNode_000', 'VarNode_001']
        observation = self.rng.normal(size=num_obser)
        mu = self.rng.normal(size=dim)
        sigma = self.rng.uniform(size=dim)

        var = VarNode(dim, node_coef=0)
        go_var = GOVarNode(observation)
        factor = GOFactorNode(connected_names, mu, sigma)

        go_graph = GOGraph()
        go_graph.add_varnode(var)
        go_graph.add_varnode(go_var)
        go_graph.add_factornode(factor)
        go_graph.bake()
        discrete_graph = go_graph.discrete_graph

        self.assertEqual(var, discrete_graph.get_node('VarNode_000'))

        target_var = VarNode(
            rv_dim=num_obser,
            potential=np.ones(num_obser),
            constrained_marginal=np.ones(num_obser) / num_obser, node_coef=0)
        target_var.name = 'VarNode_001'
        target_var.register_connection('FactorNode_000')
        self.assertEqual(target_var, discrete_graph.get_node('VarNode_001'))

        target_potential = []
        for i in range(dim):
            target_potential.append(norm.pdf(observation,
                                             loc=mu[i], scale=sigma[i]))
        target_factor = FactorNode(
            connected_names,
            np.array(target_potential))

        self.assertEqual(
            target_factor,
            discrete_graph.get_node('FactorNode_000'))



from cbp.node import FactorNode
from .base_hmmsim_builder import BaseHMMSimBuilder


class HMMSimBuilder(BaseHMMSimBuilder):
    def __init__(self, length, simulator, policy,
                 random_seed=1):
        super().__init__(length, simulator, policy, random_seed)

    def add_constrained_node(self, probability=None):
        prob = self.simulator.get_fix_margin(
            self.cnt_constrained_node)
        self.cnt_constrained_node += 1
        return super().add_constrained_node(prob)

    def add_emission(self, name_list):
        potential = self.simulator.get_emission_potential()
        factornode = FactorNode(name_list, potential)
        self.graph.add_factornode(factornode)
        return factornode

    def compare_acc(self):
        return super().compare_acc(self.graph)



import unittest

import numpy as np
from cbp.node import FactorNode, VarNode
from cbp.graph import GraphModel


class TestGraph(unittest.TestCase):
    def setUp(self):
        self.graph = GraphModel()

        # init varnode
        self.varnode_1 = VarNode(2)
        self.graph.add_varnode(self.varnode_1)

        varnode_name = [node.name for node in [self.varnode_1]]
        factor_potential = np.array(
            [3, 2]
        )
        self.factornode = FactorNode(varnode_name, factor_potential)
        self.graph.add_factornode(self.factornode)

    def test_marginal_brutal_force(self):
        self.graph.exact_marginal()
        node_equal = np.isclose(
            np.array([0.6, 0.4]),
            self.graph.get_node("VarNode_000").bfmarginal)
        self.assertTrue(all(node_equal))



from cbp.node import VarNode, FactorNode

from .base_graph import BaseGraph
from .graph_model import GraphModel


class GOGraph(BaseGraph):
    """graph with continuous observation model.
    The observations node, which only exists as leaf nodes.
    Only task: construct a discrete graph with same probability
    """

    def __init__(self):
        self.discrete_graph = None
        super().__init__()

    def bake(self):
        super().bake()
        for node in self.nodes:
            node.register_nodes(self.node_recorder)

        self.construct_discrete_graph()

    def construct_discrete_graph(self):
        self.discrete_graph = GraphModel()
        self._discrete_var()
        self._discrete_factor()
        self.discrete_graph.bake()

    def _discrete_var(self):
        for node in self.varnode_recorder.values():
            if isinstance(node, VarNode):
                self.discrete_graph.add_varnode(
                    VarNode(node.rv_dim, node.potential,
                            node.constrained_marginal))
            else:
                self.discrete_graph.add_varnode(node.discrete())

    def _discrete_factor(self):
        for node in self.factornode_recorder.values():
            if isinstance(node, FactorNode):
                self.discrete_graph.add_factornode(
                    FactorNode(node.connections, node.potential)
                )
            else:
                self.discrete_graph.add_factornode(node.discrete())

    def run_bp(self):
        self.bake()
        return self.discrete_graph.run_bp()



import numpy as np

from .base_node import BaseNode
from .var_node import VarNode


class GOVarNode(BaseNode):
    def __init__(self, bins):
        self.bins = bins
        self.is_multi = True if bins.ndim > 1 else False
        super().__init__()

    def discrete(self):
        num_sample = len(self.bins)
        return VarNode(
            num_sample,
            constrained_marginal=np.ones(num_sample) /
            num_sample)



import numpy as np
from numba import njit


def nd_multiexpand(input_data, target_shape, which_dims):
    init_shape = [1] * len(target_shape)
    expand_shape = list(target_shape)
    for i, cur_dim in enumerate(which_dims):
        assert input_data.shape[i] == target_shape[cur_dim]
        init_shape[cur_dim] = target_shape[cur_dim]
        expand_shape[cur_dim] = 1

    out = input_data.reshape(tuple(init_shape))
    return np.tile(out, tuple(expand_shape))


@njit
def nd_expand(inputdata, target_shape, expand_dim):
    """expand ndarray to target shape

    :param inputdata:
    :type inputdata: list or 1d ndarray
    :param target_shape:
    :type target_shape: tuple
    :param expand_dim: [description]
    :type expand_dim: int
    :return: expanded ndarray with `target_shape`
    :rtype: ndarray

    >>> inputdata = [1,2]
    >>> target_shape = (1,2,3)
    >>> expand_dim = 1
    >>> output = np.array([
                [1,1,1],
                [2,2,2]
            ])
    """
    assert inputdata.ndim == 1
    rtn = np.zeros(target_shape)
    for idx in np.ndindex(target_shape):
        rtn[idx] = inputdata[idx[expand_dim]]
    return rtn


@njit
def reduction_ndarray(ndarray, reduction_index):
    """reduct ndarray according to one index

    :param ndarray: [description]
    :type ndarray: ndarray
    :param reduction_index: [description]
    :type reduction_index: int
    :return: [description]
    :rtype: ndarray
    """
    rtn = np.zeros(ndarray.shape[reduction_index])
    for idx, value in np.ndenumerate(ndarray):
        rtn[idx[reduction_index]] += value
    return rtn


@njit
def batch_normal_angle(angle):
    delta_x = np.cos(angle)
    delta_y = np.sin(angle)
    return np.arctan2(delta_y, delta_x)


def empirical_marginal(traj, num_bins):
    marginal = []
    for i in range(traj.shape[1]):
        bins, _ = np.histogram(
            traj[:, i], np.arange(num_bins + 1))
        marginal.append(bins / np.sum(bins))

    return np.array(marginal).reshape(traj.shape[1], num_bins)



from dataclasses import dataclass


@dataclass
class BaseConfig:
    verbose_node_send_msg: bool = False
    verbose_engine_cnp: bool = False
    verbose_itsbp_outer: bool = False

    cnp_engine_tolerance: float = 1e-5
    cnp_engine_iteration: float = 5000000

    itsbp_outer_tolerance: float = 1e-4

    @staticmethod
    def itsbp_schedule(cnt, leaf_nodes):
        cnt += 1
        return cnt % len(leaf_nodes)


baseconfig = BaseConfig()



import numpy as np
from scipy import signal


def diagonal_potential(d_1: int, d_2: int,
                       rng: np.random.RandomState) -> np.ndarray:
    factor_potential = rng.randint(4, 6, size=(d_1, d_2)) * 1.0
    dim = np.min([d_1, d_2])
    identity = np.eye(dim)
    if rng.normal(size=1) > 1:
        identity = np.flip(identity, axis=0)
    if d_2 > d_1:
        diagonal = np.concatenate([identity, np.zeros(dim, d_2 - dim)], axis=1)
    elif d_2 < d_1:
        diagonal = np.concatenate([identity, np.zeros(d_1 - dim, dim)], axis=0)
    else:
        diagonal = identity

    #diagonal = rng.permutation(diagonal)
    diagonal_dominance = np.exp(factor_potential) + diagonal * 50000
    return diagonal_dominance / np.mean(diagonal_dominance)


def diagonal_potential_different(d_1: int, d_2: int,
                                 rng: np.random.RandomState) -> np.ndarray:
    factor_potential = rng.randint(4, 10, size=(d_1, d_2)) * 1.0
    dim = np.min([d_1, d_2])
    identity = np.eye(dim)
    if rng.normal(size=1) > 1:
        identity = np.flip(identity, axis=0)
    if d_2 > d_1:
        diagonal = np.concatenate([identity, np.zeros(dim, d_2 - dim)], axis=1)
    elif d_2 < d_1:
        diagonal = np.concatenate([identity, np.zeros(d_1 - dim, dim)], axis=0)
    else:
        diagonal = identity

    #diagonal = rng.permutation(diagonal)
    diagonal_dominance = np.exp(factor_potential) + diagonal * 50000
    return diagonal_dominance / np.mean(diagonal_dominance)


def diagonal_potential_conv(d_1: int, d_2: int,
                            rng: np.random.RandomState) -> np.ndarray:
    kernel = np.array([[0.5, 1, 0.5]])
    factor_potential = rng.randint(4, 10, size=(d_1, d_2)) * 1.0
    dim = np.min([d_1, d_2])
    identity = np.eye(dim)
    if rng.normal(size=1) > 1:
        identity = np.flip(identity, axis=0)
    if d_2 > d_1:
        diagonal = np.concatenate([identity, np.zeros(dim, d_2 - dim)], axis=1)
    elif d_2 < d_1:
        diagonal = np.concatenate([identity, np.zeros(d_1 - dim, dim)], axis=0)
    else:
        diagonal = identity

    diagonal = rng.permutation(diagonal)
    diagonal_dominance = np.exp(factor_potential) + diagonal * 50000
    diagonal_dominance /= np.mean(diagonal_dominance)

    # return diagonal_dominance / np.mean(diagonal_dominance)
    return signal.convolve2d(diagonal_dominance, kernel, mode="same")


def identity_potential(d_1: int, d_2: int,
                       rng: np.random.RandomState) -> np.ndarray:  # pylint: disable=unused-argument
    dim = np.min([d_1, d_2])
    identity = np.eye(dim)
    if d_2 > d_1:
        diagonal = np.concatenate([identity, np.zeros(dim, d_2 - dim)], axis=1)
    elif d_2 < d_1:
        diagonal = np.concatenate([identity, np.zeros(d_1 - dim, dim)], axis=0)
    else:
        diagonal = identity

    return diagonal



import numpy as np
from scipy.stats import norm, multivariate_normal

from .base_node import BaseNode
from .factor_node import FactorNode
from .go_var_node import GOVarNode


class GOFactorNode(BaseNode):
    def __init__(self, connections, loc, scale):
        assert loc.ndim == 1
        assert loc.shape == scale.shape
        self.loc = loc
        self.scale = scale
        super().__init__()
        self.connections = connections

    def _discrete_var(self, varnode):
        assert isinstance(varnode, GOVarNode)
        potenital = []
        pdf_func = multivariate_normal if varnode.is_multi else norm
        for loc, scale in zip(self.loc, self.scale):
            potenital.append(pdf_func(loc, scale).pdf(varnode.bins))
            # potenital.append(norm.pdf(varnode.bins, loc=loc, scale=scale))
        return FactorNode(self.connections, np.array(potenital))

    def discrete(self):
        for node in self.connected_nodes.values():
            if isinstance(node, GOVarNode):
                return self._discrete_var(node)

        raise RuntimeError("go factor has no GOVarNode")



import pickle
from abc import ABC, abstractmethod
from enum import Enum
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from numba import jit
from numpy.random import RandomState

from cbp.utils.np_utils import empirical_marginal


class PotentialType(str, Enum):
    """Potential Type Enum
    """
    INIT = 'Init'
    TRANSITION = 'Transition'
    EMISSION = 'Emission'


@jit(cache=True, nopython=True)
def count_joint(traj, col_num, state_num):
    idx = traj[:, col_num:col_num + 2]
    empty = np.zeros((state_num, state_num))
    for i, j in zip(idx[:, 0], idx[:, 1]):
        empty[i, j] += 1
    return empty / traj.shape[0]


class TrajData(dict):
    def __init__(self, time_step, state_num):  # pylint: disable=super-init-not-called
        self.time_step = time_step
        self.state_num = state_num

    @property
    def traj(self):
        return dict.__getitem__(self, "traj")

    @traj.setter
    def traj(self, traj_data):
        assert traj_data.shape[1] == self.time_step
        dict.__setitem__(self, "traj", traj_data)
        dict.__setitem__(self, "gt_margin", empirical_marginal(
            traj_data, self.state_num))

    def _traj_joint(self):
        record = []
        for i in range(self.time_step - 1):
            record.append(count_joint(self["traj"], i, self.state_num))
        self["gt_joint"] = record

    def cal_traj_theory(self):
        transition = self[PotentialType.TRANSITION].T
        init = self[PotentialType.INIT].reshape(-1, 1)
        state_record = []
        for _ in range(self.time_step):
            state_record.append(init.flatten())
            init = transition @ init
        self["th_margin"] = np.array(state_record)
        state_err = np.linalg.norm(self["th_margin"] -
                                   self["gt_margin"])
        self["sim_state_error"] = state_err


class TrajSimulator(ABC):
    """discrete traj simulator, abstract inference for observation

    :param ABC: [description]
    :type ABC: [type]
    """

    def __init__(self, record, random_seed, is_theorem=False):
        self.__name = None
        self.path = Path('data/sim')
        self.path.mkdir(parents=True, exist_ok=True)
        self.is_theorem = is_theorem
        self.rng = RandomState(random_seed)
        self.record = record

    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, new_name):
        self.__name = new_name
        self.path = Path(f"data/sim/{new_name}")
        self.path.mkdir(parents=True, exist_ok=True)

    def register_potential(self, ptype, potential):
        """register potential for simulation

        :param ptype: potential type
        :type ptype: PotentialType
        :param potential: [description]
        :type potential: ndarray
        """
        assert isinstance(ptype, PotentialType)

        if ptype == PotentialType.TRANSITION:
            assert np.isclose(potential.sum(axis=1), 1).all(),\
                "potential should be a conditional distribution"
            assert potential.shape == (
                self.record.state_num, self.record.state_num)
        elif ptype == PotentialType.INIT:
            assert potential.shape == (self.record.state_num,)

        self.record[ptype] = potential

    def sample_engine(self, state, dim, potential):
        """sample according to conditional prob table

        :param state: cur_state
        :type state: int
        :param dim: range of next state or observation
        :type dim: int
        :param potential: conditional prob table
        :type potential: ndarray
        :return: next state or observation
        :rtype: int
        """
        assert state < self.record.state_num
        conditional_prob = potential[state, :]
        next_state = self.rng.choice(dim, p=conditional_prob)
        return next_state

    def step(self, state):
        """do a states transition sample

        :param state: current states
        :type state: int
        :return: the next states
        :rtype: int
        """
        return self.sample_engine(state, self.record.state_num,
                                  self.record[PotentialType.TRANSITION])

    def __init_stats_sampler(self):
        return self.rng.choice(self.record.state_num,
                               p=self.record[PotentialType.INIT])

    def sample_traj(self, num_sample):
        self.record["num_sample"] = num_sample
        traj_recorder = []
        for _ in range(num_sample):
            states = []
            single_state = self.__init_stats_sampler()
            for _ in range(self.record.time_step):
                states.append(single_state)
                single_state = self.step(single_state)
            traj_recorder.append(states)

        self.record.traj = np.array(traj_recorder).reshape(num_sample, -1)

    def viz_trans_potential(self):
        axes = sns.heatmap(self.record[PotentialType.TRANSITION])
        axes.set_title("Transition Potential")
        fig = axes.get_figure()
        fig.savefig(f"{self.path}/hmm_transition.png")
        plt.close(fig)

    def viz_gt(self):
        gt_margin = self.get_hidden_margin()
        axes = sns.heatmap(gt_margin.T)
        axes.set_title("Evolution of distribution")
        axes.set_xlabel("time")
        fig = axes.get_figure()
        fig.savefig(f"{self.path}/gt.png")
        plt.close(fig)

    def get_init_potential(self):
        return self.record[PotentialType.INIT]

    def get_transition_potential(self):
        return self.record[PotentialType.TRANSITION]

    def get_hidden_margin(self, time_step=None):
        """return ground truth marginal

        :param time_step: if int then return a specific time distribution
        otherwise all distributions as matrix, defaults to None
        :type time_step: int, optional
        :return: array for single time_step or a matrix
        :rtype: ndarray
        """
        key_word = "th_margin" if self.is_theorem else "gt_margin"
        if isinstance(time_step, int):
            return self.record[key_word][time_step, :]

        return self.record[key_word]

    @abstractmethod
    def observe(self, state):
        """do a emission sample

        :param state: current states
        :type state: int
        :return: the observation of cur states
        :rtype: int
        """

    def observe_traj(self, traj):
        rtn = np.zeros_like(traj,dtype=float)
        loop_iter = np.nditer(traj, flags=['multi_index'])
        for i in loop_iter:
            rtn[loop_iter.multi_index] = self.observe(i)
        return rtn

    def sample(self, num_sample):
        self.sample_traj(num_sample)
        self.record.sensor = self.observe_traj(self.record.traj)

    def get_fix_margin(self, time_step=None):
        """return observation marginal

        :param time_step: if int then return a specific time distribution
        otherwise all distributions as matrxi, defaults to None
        :type time_step: int, optional
        :return: array for single time_step or a matrix
        :rtype: ndarray
        """
        key_word = "th_observation" if self.is_theorem else "fix_margin"
        if isinstance(time_step, int):
            return self.record[key_word][time_step, :]

        return self.record[key_word]

    def get_emission_potential(self):
        return self.record[PotentialType.EMISSION]

    def save(self):
        sim_path = f"{self.path}/sim.pkl"
        with open(sim_path, 'wb') as handle:
            pickle.dump(self, handle)

    @classmethod
    def load(cls, sim_name: str):
        """load a simulator from pkl file

        :param sim_name: name for the simulator
        :type sim_name: str
        :return: simulator
        :rtype: simulator
        """
        sim_path = f"data/sim/{sim_name}/sim.pkl"
        with open(sim_path, 'rb') as handle:
            simulator = pickle.load(handle)
        return simulator



import numpy as np
from cbp.node import FactorNode, VarNode

from .hmm_builder import HMMBuilder
from .potential_utils import identity_potential


class HMMZeroBuilder(HMMBuilder):
    def __init__(self, length, d, policy, rand_seed=1):
        super().__init__(length, d, policy, rand_seed=rand_seed)

    def add_factor(self, name_list, is_obser=False):
        factor_potential = identity_potential(
            self.node_dim, self.node_dim, self.rng)
        factornode = FactorNode(name_list, factor_potential)
        self.graph.add_factornode(factornode)
        return factornode

    def add_constrained_node(self, probability=None):
        probability = np.zeros(self.node_dim)
        probability[0] = 1
        probability[-1] = 1
        varnode = VarNode(
            self.node_dim,
            constrained_marginal=probability /
            np.sum(probability))
        self.graph.add_varnode(varnode)
        return varnode



from abc import abstractmethod
import numpy as np

from cbp.node import FactorNode
from .hmm_builder import HMMBuilder


class BaseHMMSimBuilder(HMMBuilder):
    def __init__(self, length, simulator, policy,
                 random_seed=1):
        super().__init__(length, simulator.record.state_num,
                         policy, random_seed)
        self.simulator = simulator
        self.cnt_constrained_node = 0

    def fix_initpotential(self, potential=None):
        first_node = self.graph.get_node(f"VarNode_{0:03d}")
        if potential is None:
            potential = self.simulator.get_init_potential()
        first_node.potential = potential

    def add_factor(self, name_list, is_obser=False):
        """add factor to hmm graph

        :param name_list: connected node
        :type name_list: list
        :param is_obser: is emit or not , defaults to False-- transition
        :type is_obser: bool, optional
        :return: FactorNode
        :rtype: cbp.FactorNode
        """
        if is_obser:  # emit
            return self.add_emission(name_list)

        return self.add_transition(name_list)

    def add_transition(self, name_list):
        potential = self.simulator.get_transition_potential()
        factornode = FactorNode(name_list, potential)
        self.graph.add_factornode(factornode)
        return factornode

    @abstractmethod
    def add_emission(self, name_list):
        """interface to implement
        """

    def add_branch(self, head_node=None, is_constrained=False,
                   prob=None, is_obser=False):
        if is_obser:  # emit
            return super().add_branch(head_node, True, prob, is_obser)
        return super().add_branch(head_node, False, prob, is_obser)

    def example(self, num_sample):
        self.simulator.example(num_sample)

    def compare_acc(self, graph):
        infer_marginal = []
        gt_marginal = []
        for i in range(0, self.hmm_length, 2):
            infer_marginal.append(graph.get_node(f'VarNode_{i:03d}').marginal())
            gt_marginal.append(self.simulator.get_hidden_margin(i // 2))
        return np.sum(np.abs(np.array(infer_marginal) - np.array(gt_marginal)))



import unittest

from cbp.builder import MigrSimulator


class TestMigrSim(unittest.TestCase):
    def test_process(self):
        sim = MigrSimulator(3, 3, 3, 1)
        sim.compile()
        sim.sample(10)
        sim.viz_gt()
        sim.viz_sensor()
        sim.viz_emission_potential()
        sim.viz_trans_potential()



import numpy as np
from cbp.utils.np_utils import nd_expand

from .discrete_node import DiscreteNode


class FactorNode(DiscreteNode):
    """Factor Node in factor graph.
      Add new attr:
        * ``isconstrained`` Fixed marginal or not
        * ``hat_c_ialpha`` See Norm-Product paper
        * ``last_innerparenthese_msg`` See Norm-Product paper
    """

    def __init__(self, connections, potential, coef=1):
        """except calling `~cbp.node.BaseNode` do init work
        * init last_innerparenthese_msg for cnp
        * check the connections should obey increasing order, the reason is
        defualt first dimension of potential coresponding to parent node

        :param connections: list of str, names connected variable nodes
        :type connections: list
        :param potential: 2d potential matrixx
        :type potential: ndarray
        :param coef: [description], defaults to 1
        :type coef: int, optional
        :raises RuntimeError: [description]
        """
        super().__init__(coef, potential)
        self.connections = connections
        self.last_innerparenthese_msg = {}
        self.hat_c_ialpha = {}
        self.i_alpha = {}
        self.__init_cnp()

    def __init_cnp(self):
        num_connectednode = []
        for item in self.connections:
            self.i_alpha[item] = None
            self.last_innerparenthese_msg[item] = np.ones(
                self.potential.shape)
            num_connectednode.append(int(item[-3:]))

        if any(i > j for i, j in zip(num_connectednode, num_connectednode[1:])):
            raise RuntimeError('Set the connection of factor in order')

    def check_before_run(self, node_map):
        super().check_before_run(node_map)
        self.check_dim(node_map)

    def check_dim(self, node_map):
        """ * check the discrete potential matrix dimension

        :param node_map: a map contains the var nodes
        :type node_map: map
        """
        for i, varnode_name in enumerate(self.connections):
            varnode = node_map[varnode_name]
            assert self.potential.shape[i] == varnode.rv_dim, \
                f"Dimention mismatch! At {i:02d} axis in Factor:{self.name} \
                    rv_dim:{varnode.rv_dim:02d}, \
                    potential: {self.potential.shape[i]}"

    def _check_potential(self, potential):
        """normalize the potential
        """
        return potential / np.sum(potential)

    def auto_coef(self, node_map, assign_policy=None):
        super().auto_coef(node_map, assign_policy)

        sum_i_alpha, unset_edge = self.__sum_neighbor_coef()
        if unset_edge:
            new_i_alpha = 1 - self.node_coef - sum_i_alpha
            self.set_i_alpha(unset_edge, new_i_alpha)

    def __sum_neighbor_coef(self):
        sum_i_alpha = 0
        uninit_neighbor = None
        for connected_var in self.connections:
            i_alpha = self.get_i_alpha(connected_var)
            if i_alpha is not None:
                sum_i_alpha += i_alpha
            else:
                uninit_neighbor = connected_var
        return sum_i_alpha, uninit_neighbor

    def get_i_alpha(self, connection_name):
        return self.i_alpha[connection_name]

    def set_i_alpha(self, connection_name, value):
        self.i_alpha[connection_name] = value

    def cal_cnp_coef(self):
        self.coef_ready = True

        for item in self.connections:
            hat_c_ialpha = self.node_coef + self.i_alpha[item]
            assert hat_c_ialpha != 0
            self.hat_c_ialpha[item] = hat_c_ialpha

    def get_hat_c_ialpha(self, node_name):
        if self.coef_ready:
            return self.hat_c_ialpha[node_name]
        return None

    def get_varnode_extra_term(self, node_name):
        """
        Norm-Product Belief Propagation, n_{i -> alpha} second term
        This term is always 1 in stardard bp
        """
        if node_name not in self.last_innerparenthese_msg:
            raise RuntimeError(
                f"{node_name} do not have previous msg sent by {self.name}")

        if abs(self.i_alpha[node_name]) < 1e-5 and self.node_coef == 1:
            return np.ones_like(
                self.last_innerparenthese_msg[node_name])

        # TODO when the a^x, a = 0, it has some problem
        coef_exp = -1.0 * \
            self.i_alpha[node_name] / self.hat_c_ialpha[node_name]
        base = self.last_innerparenthese_msg[node_name]
        value = np.power(base, coef_exp)
        return value

    def make_message(self, recipient_node):
        assert recipient_node.name in self.connections
        if len(self.connections) == 1:
            self.last_innerparenthese_msg[recipient_node.name] = self.potential
            return self.summation(self.potential, recipient_node)

        product_out = self.cal_inner_parentheses(recipient_node)

        with np.errstate(divide='raise'):
            hat_c_ialpha = self.hat_c_ialpha[recipient_node.name]
            log_media = 1.0 / hat_c_ialpha * \
                np.log(np.clip(product_out, 1e-12, None))
            product_out_power = np.exp(log_media)
            return np.power(
                self.summation(
                    product_out_power,
                    recipient_node),
                hat_c_ialpha)

    def cal_bethe(self, margin):
        clip_potential = np.clip(self.potential, 1e-12, None)
        return np.sum(margin * np.log(margin / clip_potential))

    def marginal(self):
        message_val = np.array([message.val for message in self.latest_message])
        prod_messages = np.prod(message_val, axis=0)
        product_out = np.multiply(self.potential, prod_messages)
        unormalized = np.power(product_out, 1.0 / self.node_coef)
        return unormalized / np.sum(unormalized)

    def cal_inner_parentheses(self, recipient_node):
        latest_message = self.latest_message
        filtered_message = [message for message in latest_message
                            if not message.sender.name == recipient_node.name]

        message_val = np.array([message.val for message in filtered_message])

        prod_messages = np.prod(message_val, axis=0)

        product_out = np.multiply(self.potential, prod_messages)
        self.last_innerparenthese_msg[recipient_node.name] = product_out
        return product_out

    def store_message(self, message):
        assert message.val.shape == self.potential.shape, \
            f"From {message.sender.name} to {self.name} shape mismatch, \
                expected {self.potential.shape}, received {message.val.shape}"
        super().store_message(message)

    def reformat_message(self, message):
        potential_dims = self.potential.shape
        states = message.val
        which_dim = self.connections.index(message.sender.name)

        return nd_expand(states, potential_dims, which_dim)

    def summation(self, potential, node):
        potential_dim = potential.shape
        node_index = self.connections.index(node.name)
        assert potential_dim[node_index] == node.rv_dim
        return potential.sum(
            tuple(j for j in range(potential.ndim) if j != node_index))

    def __eq__(self, value):
        if isinstance(value, FactorNode):
            return super().__eq__(value)

        return False

    def plot(self, graph):
        graph.add_node(self.name, color='green')
        for var_name in self.connections:
            graph.add_edge(self.name, var_name)



import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from cbp.utils.np_utils import empirical_marginal

from .base_simulator import PotentialType, TrajData, TrajSimulator


class DiscreteObserData(TrajData):
    def __init__(self, time_step, state_num, obser_num):
        super().__init__(time_step, state_num)
        self.obser_num = obser_num

    @property
    def sensor(self):
        return dict.__getitem__(self, "sensor")

    @sensor.setter
    def sensor(self, sensor_data):
        assert sensor_data.shape[1] == self.time_step
        dict.__setitem__(self, "sensor", sensor_data)
        dict.__setitem__(self, "fix_margin", empirical_marginal(
            self["sensor"], self.obser_num))

    def cal_theory(self, verbose):
        self.cal_traj_theory()
        observation_record = []
        emis = self[PotentialType.EMISSION].T
        traj = self["th_margin"]
        for i in range(self.time_step):
            obs = emis @ traj[i].reshape(-1, 1)
            observation_record.append(obs.flatten())
        self["th_observation"] = np.array(observation_record)
        obser_err = np.linalg.norm(self["th_observation"] -
                                   self["fix_margin"])
        self["sim_obser_error"] = obser_err
        state_err = self["sim_state_error"]
        if verbose:
            print(f"Sim state err: {state_err}")
            print(f"Sim obser err: {obser_err}")


class HMMSimulator(TrajSimulator):
    def __init__(self, time_step, dim_states, dim_observations, random_seed, is_theorem=False):  # pylint: disable=too-many-arguments
        record = DiscreteObserData(time_step, dim_states, dim_observations)
        super().__init__(record, random_seed, is_theorem)

    def register_potential(self, ptype, potential):
        if ptype == PotentialType.EMISSION:
            assert np.isclose(potential.sum(axis=1), 1).all(),\
                "potential should be a conditional distribution"
            assert potential.shape == (
                self.record.state_num, self.record.obser_num)
        return super().register_potential(ptype, potential)

    def observe(self, state):
        return self.sample_engine(state, self.record.obser_num,
                                  self.record[PotentialType.EMISSION])

    def viz_emission_potential(self):
        axes = sns.heatmap(self.record[PotentialType.EMISSION])
        axes.set_title("Emission Potential")
        fig = axes.get_figure()
        fig.savefig(f"{self.path}/hmm_emission.png")
        plt.close(fig)

    def viz_sensor(self):
        sensor = self.get_fix_margin()
        axes = sns.heatmap(sensor.T)
        axes.set_title("Evolution of distribution")
        axes.set_xlabel("time")
        fig = axes.get_figure()
        fig.savefig(f"{self.path}/sensor.png")
        plt.close(fig)

    def get_precious(self, time_step=None, verbose=False):
        """return precious marginal

        :param time_step: if int then return a specific time distribution
        otherwise all distributions as matrix, defaults to None
        :type time_step: int, optional
        :param verbose: whether or not ouput difference between simulation and theorical margin
        :return: array for single time_step or a matrix
        :rtype: ndarray
        """
        if "th_margin" not in self.record:
            self.record.cal_theory(verbose)

        if isinstance(time_step, int):
            return self.record["th_margin"][time_step, :]

        return self.record["th_margin"]



from .base_builder import BaseBuilder
from .base_simulator import PotentialType
from .go_simulator import GOHMMSimulator
from .hmm_builder import HMMBuilder
from .hmm_engine import HMMEngine
from .hmm_sim_builder import HMMSimBuilder
from .hmm_simulator import HMMSimulator
from .hmm_zero_builder import HMMZeroBuilder
from .line_builder import LineBuilder
from .migr_simulator import MigrSimulator
from .star_builder import StarBuilder
from .tree_builder import TreeBuilder
from .wifi_hmm_builder import WifiHMMBuilder
from .wifi_simulator import WifiSimulator
from .hmm_gosim_builder import HMMGOSimBuilder

__all__ = [
    "HMMBuilder",
    "BaseBuilder",
    "LineBuilder",
    "StarBuilder",
    "TreeBuilder",
    "HMMZeroBuilder",
    "WifiSimulator",
    "MigrSimulator",
    "WifiHMMBuilder",
    "HMMSimulator",
    "PotentialType",
    "HMMSimBuilder",
    "HMMEngine",
    "GOHMMSimulator",
    "HMMGOSimBuilder"
]



from cbp.node.base_node import BaseNode
from cbp.node.var_node import VarNode
from cbp.node.factor_node import FactorNode
from cbp.node.go_factor_node import GOFactorNode
from cbp.node.go_var_node import GOVarNode

__all__ = [
    "BaseNode",
    "VarNode",
    "FactorNode",
    "GOVarNode",
    "GOFactorNode"
]



import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns


class MigrVisualizer():
    def __init__(self, d_row, d_col):
        """initial vis. TopLeft is origin point.

        :param d_row: grid height
        :type d_row: int
        :param d_col: grid width
        :type d_col: grid width
        """
        self.d_row = d_row
        self.d_col = d_col
        self.states_num = d_row * d_col  # total states

    def potential_heatmap(self, data, **kwargs):
        """plot potential heatmap and save png.
        2 reserved keys:

        * ``title`` figure title
        * ``path`` figure path prefix

        :param data: i-th row represents the potential of d state
        :type data: ndarray
        """
        for i in range(self.states_num):
            distribution = data[i, :].reshape(self.d_row, self.d_col)
            axes = sns.heatmap(distribution)
            axes.set_title(f"{self.ind2rowcol(i)}_{kwargs['title']}")
            fig = axes.get_figure()
            fig.savefig(f"{kwargs['path']}_{i}_step.png")
            plt.close(fig)

    # FIXME: CHANGE API
    def visualize_location(self, xx, yy, xy_size, **kwargs):
        """plot grid location distribution. Origin point is in DownLeft.
        3 reservered key in kwargs:
        * ``fig_name``: savefig
        * ``xlabel``: xlabel
        * ``ylabel``

        :param xx: scatter plot x
        :type xx: ndarray
        :param yy: scatter plot y
        :type yy: ndarray
        :param xy_size: at position (x,y) the number of particles
        :type xy_size: ndarray
        """
        cmap = sns.cubehelix_palette(dark=.3, light=.8, as_cmap=True)
        # FIXME: remove the comments. use the quickplot to refactor code here
        with sns.axes_style("whitegrid"):
            axes = sns.scatterplot(x=xx, y=yy, size=xy_size,
                                   sizes=(0, np.sqrt(np.max(xy_size)) /
                                          np.sqrt(np.sum(xy_size)) * 5000),
                                   palette=cmap)

            axes.set_xlim([0 - 1, self.d_col + 1])
            axes.set_ylim([0 - 1, self.d_row + 1])

            axes.set_xticks(np.arange(1, self.d_col + 1, 3))
            axes.set_xticks(np.arange(1, self.d_col + 1), minor=True)
            axes.set_yticks(np.arange(1, self.d_row + 1, 3))
            axes.set_yticks(np.arange(1, self.d_col + 1), minor=True)
            axes.set_yticklabels([])
            axes.set_xticklabels([])

            if "ylabel" in kwargs:
                axes.set_ylabel(kwargs['ylabel'])
            if "xlabel" in kwargs:
                axes.set_xlabel(kwargs['xlabel'])

            axes.legend_.remove()
            axes.grid(which='both')
            axes.grid(which='minor', alpha=0.2)
            axes.grid(which='major', alpha=0.5)
            axes.yaxis.label.set_size(32)
            axes.xaxis.label.set_size(32)
            plt.savefig(kwargs["fig_name"], pad_inches=-4)
            plt.close()

    def migration(self, data, **kwargs):
        """draw migration figure, heat map distribution
        3 reservered key in kwargs:
        * ``fig_name``: savefig
        * ``xlabel``: xlabel
        * ``ylabel``

        :param data: [i,j] record i-th particle position at j timestamp
        """
        _, time_length = data.shape
        for i in range(time_length):
            locations = data[:, i]  # i timestamp distribution
            bins, _ = np.histogram(locations, np.arange(self.states_num + 1))

            if 'ylabel' in kwargs:
                kwargs['ylabel'] = f"t={i}"
            kwargs['fig_name'] = f"{kwargs['path']}_{i}_step.png"

            self.visualize_map_bins(bins, **kwargs)

    def visualize_map_bins(self, bins, **kwargs):
        """converts the statistics bins data to the map figure.
        3 reservered key in kwargs:
        * ``fig_name``: savefig
        * ``xlabel``: xlabel
        * ``ylabel``

        :param bins: every element represents how many particles in the cell
        :type bins: list or ndarray
        """
        xx = []
        yy = []
        xy_cnt = []
        for xy, cnt in enumerate(bins):
            if cnt > 0:
                row, col = self.ind2rowcol(xy)
                xx.append(col)
                yy.append(row)
                xy_cnt.append(int(cnt))
        self.visualize_location(xx, yy, xy_cnt, **kwargs)

    def ind2rowcol(self, index):
        index = np.array(index).astype(np.int64)
        row = (index / self.d_col).astype(np.int64)
        col = index % self.d_col
        return row, col



import warnings

try:
    import pygraphviz  # noqa
except BaseException:
    pygraphviz = None


class BaseGraph():
    """
    base graph type, responsible for basic graph manipulation.
        * register var node and factor node
        * delete node
        * access node
    """

    def __init__(self):
        self.varnode_recorder = {}
        self.factornode_recorder = {}

        self.leaf_nodes = []
        self.nodes = []
        self.node_recorder = {}
        self.cnt_varnode = 0
        self.cnt_factornode = 0

    def add_varnode(self, node):
        """add one variable node type to this graph, idx follow the increasing
        order

        :param node: one var node
        :type node: var node
        :return: name of var node
        :rtype: str
        """
        varnode_name = f"VarNode_{self.cnt_varnode:03d}"
        node.format_name(varnode_name)
        self.varnode_recorder[varnode_name] = node

        self.cnt_varnode += 1
        return varnode_name

    def add_factornode(self, factornode):
        """add one factor node to the graph
        Do the following tasks

        *set factornode name attr
        * add node to recorder
        * set connections
        * set parent relation

        :param factornode: one factor node
        :return: name of factor node
        :rtype: str
        """
        factornode_name = f"FactorNode_{self.cnt_factornode:03d}"
        factornode.format_name(factornode_name)
        self.factornode_recorder[factornode_name] = factornode

        self.__register_connection(factornode)
        self.__set_parent(factornode)

        self.cnt_factornode += 1
        return factornode_name

    def __register_connection(self, factornode):
        for varnode_name in factornode.connections:
            varnode = self.varnode_recorder[varnode_name]
            varnode.register_connection(factornode.name)

    def __set_parent(self, factornode):
        connections = factornode.connections
        factornode.parent = self.varnode_recorder[connections[0]]
        for varnode_name in connections[1:]:
            varnode = self.varnode_recorder[varnode_name]
            varnode.parent = factornode

    def bake(self):
        self.init_node_list()

    def init_node_list(self):
        factors = list(self.factornode_recorder.values())
        variables = list(self.varnode_recorder.values())
        # in Norm-Product, run factor message first
        self.nodes = factors + variables
        self.leaf_nodes = [
            node for node in self.nodes if len(node.connections) == 1]
        self.node_recorder = {
            **self.factornode_recorder,
            **self.varnode_recorder}

    def get_root(self):
        if len(self.leaf_nodes) == 0:
            raise RuntimeError("No leaf node find in the graph")
        return self.leaf_nodes[0]

    def get_node(self, name_str):
        """access to the node in the graph

        :param name_str: name of node
        :type name_str: string
        :raises RuntimeError: not in the graph
        """
        if name_str in self.varnode_recorder:
            return self.varnode_recorder[name_str]
        elif name_str in self.factornode_recorder:
            return self.factornode_recorder[name_str]
        else:
            raise RuntimeError(f"{name_str} is illegal, not in this graph")

    def delete_node(self, name_str):
        """delete node from graph, needs to check following

        * clear connections of connected nodes
        * delete from the various recorders
        * call init_node_list

        :param name_str: name of delete node
        :type name_str: str
        :raises RuntimeError: node name not in this graph
        """
        if name_str not in self.node_recorder:
            raise RuntimeError(f"{name_str} is illegal, not in this graph")
        target_node = self.node_recorder[name_str]
        if 'var' in name_str.lower():
            warnings.warn(f"Delete {name_str}, may have a suspend factor node")
        self._delete_node_connection(target_node)
        self._delete_node_recorder(target_node)

    def _delete_node_connection(self, target_node):
        for connected_name in target_node.connections:
            connected_node = self.node_recorder[connected_name]
            if connected_node.parent is target_node:
                connected_node.parent = None
            connected_node.connections.remove(target_node.name)

            # clear map
            if len(connected_node.connections) == 0:
                self._delete_node_recorder(connected_node)

    def _delete_node_recorder(self, node):
        target_map = self.varnode_recorder if 'var' in node.name.lower(
        ) else self.factornode_recorder
        del target_map[node.name]

    def plot(self, png_name='file.png'):
        """plot the graph through graphviz

            * delegate node plot to different node type

        :param png_name: name of figure, defaults to 'file.png'
        :type png_name: str, optional
        :raises ValueError: pygraphviz not installed
        """
        if pygraphviz is not None:
            graph = pygraphviz.AGraph(directed=False)
            for node in self.varnode_recorder.values():
                node.plot(graph)

            for node in self.factornode_recorder.values():
                node.plot(graph)

            graph.layout(prog='neato')
            graph.draw(png_name)
        else:
            raise ValueError("must have pygraphviz installed for visualization")



import unittest
from cbp.builder import MigrSimulator, HMMSimBuilder
from cbp.graph import bp_policy


class TestITSBPHMM(unittest.TestCase):
    def test_acc(self):
        sim = MigrSimulator(15, 15, 15, 1)
        sim.compile()
        sim.sample(10)
        builder = HMMSimBuilder(15, sim, bp_policy)
        graph = builder()
        graph.bake()
        builder.fix_initpotential()
        graph.first_belief_propagation()
        print(0, builder.compare_acc())
        for i in range(20):
            graph.itsbp_outer_loop()
            print(i, builder.compare_acc())



from .message import Message
from . import np_utils
from .event_utils import engine_loop, compare_marginals, diff_max_marginals

__all__ = [
    "Message",
    "np_utils",
    "engine_loop",
    "compare_marginals",
    "diff_max_marginals"
]



from . import coef_policy
from .base_graph import BaseGraph
from .coef_policy import bp_policy
from .discrete_graph import DiscreteGraph
from .go_graph import GOGraph
from .graph_model import GraphModel

__all__ = [
    "BaseGraph",
    "DiscreteGraph",
    "GraphModel",
    "GOGraph",
    "bp_policy",
    "coef_policy"
]



def avg_policy(node, node_map):  # pylint: disable=unused-argument
    return 1.0 / len(node_map)


def factor_policy(node, node_map):
    cnt_var = 0
    cnt_factor = 0
    for name, _ in node_map.items():
        if 'VarNode' in name:
            cnt_var += 1
        elif 'FactorNode' in name:
            cnt_factor += 1
        else:
            raise RuntimeError(f"{name} is wrong")

    if node.__class__.__name__ == 'VarNode':
        coef = 0
    elif node.__class__.__name__ == "FactorNode":
        coef = 1.0 / cnt_factor
    return coef


def bp_policy(node, node_map):  # pylint: disable=unused-argument
    if node.__class__.__name__ == 'VarNode':
        coef = 1.0 - len(node.connections)
    elif node.__class__.__name__ == 'FactorNode':
        coef = 1.0
    return coef


def crazy_policy(node, node_map):
    cnt_var = 0
    cnt_factor = 0
    propotion_coef = 2.0
    for name, _ in node_map.items():
        if 'VarNode' in name:
            cnt_var += 1
        elif 'FactorNode' in name:
            cnt_factor += 1
        else:
            raise RuntimeError(f"{name} is wrong")

    if node.__class__.__name__ == 'VarNode':
        coef = 1.0 - len(node.connections) - propotion_coef * \
            (cnt_factor + cnt_var) / cnt_var
    elif node.__class__.__name__ == "FactorNode":
        coef = 1 + propotion_coef * (cnt_factor + cnt_var) / cnt_factor
    return coef



from cbp.node import GOFactorNode, GOVarNode
from cbp.graph import GOGraph
from .base_hmmsim_builder import BaseHMMSimBuilder


class HMMGOSimBuilder(BaseHMMSimBuilder):
    def __init__(self, length, simulator, policy, random_seed=1):
        super().__init__(length, simulator, policy, random_seed=random_seed)

    def _create_graph(self):
        return GOGraph()

    def add_constrained_node(self, probability=None):
        bins = self.simulator.get_fix_margin(
            self.cnt_constrained_node)
        self.cnt_constrained_node += 1
        node = GOVarNode(bins)
        self.graph.add_varnode(node)
        return node

    def add_emission(self, name_list):
        potential = self.simulator.get_emission_potential()
        factornode = GOFactorNode(name_list, **potential)
        self.graph.add_factornode(factornode)
        return factornode

    def compare_acc(self):
        return super().compare_acc(self.graph.discrete_graph)



import uuid
from abc import ABC
import numpy as np


class BaseNode(ABC):
    """All kinds node must inherit :class:`~cbp.node.BaseNode`. It
    provide the basic register connection functions. Important attr.
        * ``name`` str: id for the node
        * ``connections`` list of str: a list of connected node names
        * ``connected_nodes`` map of nodes: name -> node
        * ``parent`` traverse purpose
    """

    def __init__(self) -> None:
        """Initialize default attr
        """
        self.name = str(uuid.uuid4())
        self.node_degree = 0
        self.parent = None
        self.connections = []
        self.connected_nodes = {}

    def __str__(self):
        return f"{self.name}"

    def __repr__(self):
        return self.__str__()

    def format_name(self, name):
        self.name = name

    def check_before_run(self, node_map):
        for item in self.connections:
            assert item in node_map, f"{self.name} has a connection {item}, \
                                                    which is not in node_map"

    def register_connection(self, node_name):
        assert isinstance(node_name, str)
        self.node_degree += 1
        self.connections.append(node_name)

    def register_nodes(self, node_map):
        """initial `connected_nodes`. After build a graph, then do this work

        :param node_map: all nodes in a graph. name -> node
        :type node_map: dict
        :raises IOError: the input graph structure is wrong
        """
        for item in self.connections:
            if item in node_map:
                self.connected_nodes[item] = node_map[item]
            else:
                raise IOError(f"connection of {item} of {self.name} \
                                do not appear in the node_map")

    def search_node_index(self, node_name):
        return self.connections.index(node_name)

    def plot(self, graph):
        graph.add_node(self.name, color='blue', style='bold')

    def __eq__(self, value):
        flag = []
        flag.append(set(self.connections) == set(value.connections))
        flag.append(self.node_degree == value.node_degree)
        if np.sum(flag) == len(flag):
            return True

        return False



from .base_builder import BaseBuilder


class StarBuilder(BaseBuilder):
    def __init__(self, num_node, d, policy, rand_seed):
        self.num_node = num_node
        super().__init__(d, policy, rand_seed)

    def init_graph(self):
        center_node = "VarNode_000"
        self.add_trivial_node()

        for _ in range(self.num_node - 1):
            self.add_branch(head_node=center_node, is_constrained=True)



import unittest
import numpy as np
from cbp.graph import DiscreteGraph
from cbp.node import VarNode, FactorNode


class TestDiscreteGraph(unittest.TestCase):
    def setUp(self):
        self.graph = DiscreteGraph()
        self.dim = 4
        self.rng = np.random.RandomState(1)

    def test_add_varnode(self):
        var1 = VarNode(self.dim)
        var2 = VarNode(self.dim)
        self.graph.add_varnode(var1)
        self.graph.add_varnode(var2)
        self.graph.init_node_list()

        self.assertEqual(self.graph.cnt_varnode, 2)
        self.assertEqual(id(self.graph.get_node('VarNode_000')), id(var1))

        c_var1 = VarNode(
            self.dim, constrained_marginal=np.ones(
                self.dim) / self.dim)

        self.graph.add_varnode(c_var1)
        self.graph.init_node_list()
        self.assertEqual(self.graph.constrained_names, ['VarNode_002'])

    def test_delete_var(self):
        self.test_add_varnode()
        self.graph.delete_node('VarNode_000')
        self.graph.init_node_list()
        self.assertNotIn('VarNode_000', self.graph.varnode_recorder)

    def test_add_factor_node(self):
        self.test_delete_var()

        factor = FactorNode(['VarNode_001', 'VarNode_002'],
                            potential=self.rng.uniform(size=(self.dim, self.dim)))
        self.graph.add_factornode(factor)
        self.graph.init_node_list()
        self.assertEqual(id(self.graph.get_node('FactorNode_000')), id(factor))
        self.graph.delete_node('FactorNode_000')
        self.assertEqual(len(self.graph.factornode_recorder), 0)



import unittest
import numpy as np
from cbp.builder import GOHMMSimulator, HMMGOSimBuilder, PotentialType
from cbp.graph import bp_policy


def construct_init(dim):
    init = np.array([(dim - i) for i in range(dim)])
    return init / np.sum(init)


def construct_shift_trans(dim):
    ones = np.eye(dim)
    shift = np.hstack([ones[:, 1:], ones[:, 0:1]])
    return shift


def construct_gaussian_emission(dim):
    loc = 5 * np.arange(dim)
    scale = [1] * dim
    return {"loc": np.array(loc), "scale": np.array(scale)}


class TestITSBPHMM(unittest.TestCase):

    def test_acc(self):
        dim = 5
        length = 4
        sim = GOHMMSimulator(length, dim, 1)
        sim.register_potential(PotentialType.INIT, construct_init(dim))
        sim.register_potential(PotentialType.TRANSITION,
                               construct_shift_trans(dim))
        sim.register_potential(PotentialType.EMISSION,
                               construct_gaussian_emission(dim))
        sim.sample(2000)
        builder = HMMGOSimBuilder(length, sim, bp_policy)

        graph = builder()
        builder.fix_initpotential()
        graph.bake()

        discrete_graph = graph.discrete_graph
        discrete_graph.first_belief_propagation()
        for i in range(20):
            print(i, discrete_graph.get_node("VarNode_000").marginal(),
                  builder.compare_acc())
            discrete_graph.itsbp_outer_loop()



import unittest

import numpy as np
from cbp.utils.np_utils import nd_expand, nd_multiexpand


class TestNpUtils(unittest.TestCase):
    def test_nd_expand(self):
        inputdata = np.array([1, 2])
        target_shape = (1, 2, 3)
        expand_dim = 1
        output = nd_expand(inputdata, target_shape, expand_dim)
        target = np.array([
            [1, 1, 1],
            [2, 2, 2]
        ])
        equal = np.isclose(output, target)
        self.assertTrue(equal.all())

    def test_nd_multiexpand(self):
        in_ = np.array([[1, 2]])
        out_shape = (1, 2, 3)
        dims = (0, 1)
        out_ = nd_multiexpand(in_, out_shape, dims)
        target = np.array([
            [1, 1, 1],
            [2, 2, 2]
        ])
        equal = np.isclose(out_, target)
        self.assertTrue(equal.all())



import numpy as np

from .base_simulator import PotentialType, TrajData, TrajSimulator


class ContinuousObserData(TrajData):
    @property
    def sensor(self):
        return dict.__getitem__(self, "fix_margin")

    @sensor.setter
    def sensor(self, sensor_data):
        assert sensor_data.shape[1] == self.time_step
        dict.__setitem__(self, "fix_margin", np.swapaxes(sensor_data, 0, 1))


class GOHMMSimulator(TrajSimulator):
    def __init__(self, time_step, dim_states, random_seed):
        record = ContinuousObserData(time_step, dim_states)
        super().__init__(record, random_seed, is_theorem=False)

    def register_potential(self, ptype, potential):
        if ptype == PotentialType.EMISSION:
            assert potential["loc"].shape == (self.record.state_num,)
            assert potential["scale"].shape == (self.record.state_num,)

        super().register_potential(ptype, potential)

    def observe(self, state):
        return self.rng.normal(loc=self.record[PotentialType.EMISSION]["loc"][state],
                               scale=self.record[PotentialType.EMISSION]["scale"][state])


class GMOHMMSimulator(TrajSimulator):
    def __init__(self, record, random_seed, is_theorem=False):
        super().__init__(record, random_seed, is_theorem=is_theorem)

    def register_potential(self, ptype, potential):
        if ptype == PotentialType.EMISSION:
            if potential["loc"].ndim > 1:
                setattr(self, "obser_dim", potential["loc"].shape[1])
            else:
                setattr(self, "obser_dim", 1)
            assert potential["loc"].shape[0] == self.record.state_num
            assert potential["scale"].shape[0] == self.record.state_num

        super().register_potential(ptype, potential)

    def observe_traj(self, traj):
        rtn = np.zeros_like((traj.shape[0], traj.shape[1], self.obser_dim),
                            traj, dtype=float)
        loop_iter = np.nditer(traj, flags=['multi_index'])
        for i in loop_iter:
            rtn[loop_iter.multi_index] = self.observe(i)
        return rtn

    def observe(self, state):
        return self.rng.multivariate_normal(mean=self.record[PotentialType.EMISSION]["loc"][state],
                                            cov=self.record[PotentialType.EMISSION]["scale"][state])



from cbp.utils import (compare_marginals, diff_max_marginals,
                       engine_loop)
from cbp.configs.base_config import baseconfig
from cbp.node import VarNode

from .coef_policy import bp_policy
from .graph_utils import itsbp_inner_loop, find_link
from .discrete_graph import DiscreteGraph


class GraphModel(DiscreteGraph):
    def __init__(self, coef_policy=bp_policy, config=baseconfig):
        super().__init__(coef_policy=coef_policy)
        self.cfg = config
        self.itsbp_outer_cnt = 0

    def add_varnode(self, node):
        """ add check node type and call parent checker

        :param node: one VarNode
        :type node: VarNode
        :return: name of varnode
        :rtype: str
        """
        assert isinstance(node, VarNode)
        return super().add_varnode(node)

    def bake(self):
        super().bake()
        for node in self.nodes:
            node.cal_cnp_coef()

    def run_cnp(self):
        self.bake()
        return self.norm_product_bp()

    def run_bp(self):
        if self.coef_policy != bp_policy:  # pylint: disable=comparison-with-callable
            self.coef_policy = bp_policy
        self.bake()
        return self.itsbp()

    def norm_product_bp(self, error_fun=None):
        if error_fun is None:
            error_fun = diff_max_marginals
        self.first_belief_propagation()
        return self.engine_loop(
            max_iter=self.cfg.cnp_engine_iteration,
            engine_fun=self.parallel_message,
            tolerance=self.cfg.cnp_engine_tolerance,
            error_fun=error_fun,
            isoutput=self.cfg.verbose_engine_cnp)

    def engine_loop(  # pylint: disable= too-many-arguments
            self,
            engine_fun,
            max_iter=5000000,
            tolerance=1e-2,
            error_fun=None,
            isoutput=False):
        if error_fun is None:
            error_fun = compare_marginals

        epsilons, step, timer = engine_loop(
            engine_fun=engine_fun,
            max_iter=max_iter,
            tolerance=tolerance,
            error_fun=error_fun,
            meassure_fun=self.export_convergence_marginals,
            isoutput=isoutput)

        return epsilons, step, timer

    def itsbp(self):
        """run sinkhorn or iterative scaling inference

        :return: [description]
        :rtype: [type]
        """
        self.first_belief_propagation()
        return self.engine_loop(self.itsbp_outer_loop,
                                tolerance=self.cfg.itsbp_outer_tolerance,
                                error_fun=diff_max_marginals,
                                isoutput=self.cfg.verbose_itsbp_outer)

    def its_next_looplink(self):
        target_node = self.leaf_nodes[self.itsbp_outer_cnt]

        next_node = self.leaf_nodes[(
            self.itsbp_outer_cnt + 1) % len(self.leaf_nodes)]

        self.itsbp_outer_cnt = self.cfg.itsbp_schedule(
            self.itsbp_outer_cnt, self.leaf_nodes)
        return target_node, find_link(target_node, next_node)

    def itsbp_outer_loop(self):
        for _ in range(len(self.leaf_nodes)):
            _, loop_link = self.its_next_looplink()
            itsbp_inner_loop(loop_link, self.cfg.verbose_node_send_msg)

    def parallel_message(self, run_constrained=True):
        for target_var in self.varnode_recorder.values():
            # sendind in messages from factors
            target_var.sendin_message(self.cfg.verbose_node_send_msg)

            if run_constrained or (not target_var.isconstrained):
                target_var.sendout_message(self.cfg.verbose_node_send_msg)



import unittest

import numpy as np
from cbp.node import GOFactorNode, GOVarNode


class TestGOVarNode(unittest.TestCase):
    def setUp(self):
        self.rng = np.random.RandomState(1)

    def test_var_construct(self):
        node = GOVarNode(self.rng.normal(size=(10)))

    def test_factor_construct(self):
        connected_names = ['VarNode_000', 'VarNode_001']
        node = GOFactorNode(connected_names,
                            loc=self.rng.normal(size=10),
                            scale=self.rng.uniform(size=10))



import unittest
import numpy as np
from cbp.node import VarNode, FactorNode


class TestDiscreteNode(unittest.TestCase):
    def setUp(self):
        self.rng = np.random.RandomState(1)

    def test_var_equal(self):
        dim = 10
        _potential = self.rng.uniform(size=dim)
        node = VarNode(rv_dim=dim, potential=_potential)
        target_node = VarNode(rv_dim=dim, potential=_potential)
        self.assertEqual(node, target_node)

        _target_node = VarNode(dim, self.rng.uniform(size=dim))
        self.assertNotEqual(node, _target_node)

    def test_constrain_var_equal(self):
        dim = 10
        _potential = self.rng.uniform(size=dim)
        node = VarNode(dim, _potential, _potential / np.sum(_potential))
        target_node = VarNode(dim, _potential, _potential / np.sum(_potential))
        self.assertEqual(node, target_node)

        new_p = self.rng.uniform(size=dim)
        _target_node = VarNode(dim, _potential, new_p / np.sum(new_p))
        self.assertNotEqual(node, _target_node)

        _target_node = VarNode(dim, new_p, _potential / np.sum(_potential))
        self.assertNotEqual(node, _target_node)

    def test_factor_equal(self):
        dim = (4, 10)
        _potential = self.rng.uniform(size=dim)

        node = FactorNode(['VarNode_000', 'VarNode_001'], _potential)
        _target = FactorNode(['VarNode_000', 'VarNode_001'], _potential)
        self.assertEqual(node, _target)

        new_p = self.rng.uniform(size=dim)
        _target = FactorNode(['VarNode_000', 'VarNode_001'], new_p)
        self.assertNotEqual(node, _target)

        _target = FactorNode(
            ['VarNode_000', 'VarNode_001', 'VarNode_002'], _potential)
        self.assertNotEqual(node, _target)



import numpy as np
from cbp.utils.np_utils import nd_expand

from .discrete_node import DiscreteNode


class VarNode(DiscreteNode):
    """Variable Node in Factor graph

        Add new attr:
        * ``isconstrained`` Fixed marginal or not
        * ``hat_c_i`` See Norm-Product paper
    """

    def __init__(self, rv_dim, potential=None,
                 constrained_marginal=None, node_coef=0):
        self.rv_dim = rv_dim
        self.hat_c_i = None
        super().__init__(node_coef, potential)

        if constrained_marginal is None:
            self.isconstrained = False
        else:
            assert constrained_marginal.shape[0] == rv_dim
            assert abs(np.sum(constrained_marginal) - 1) < 1e-6
            self.isconstrained = True
            constrained_marginal = np.clip(constrained_marginal, 1e-12, None)
        self.constrained_marginal = constrained_marginal

    def _check_potential(self, potential):
        """deal None case
        """
        if potential is None:
            return np.ones([self.rv_dim])
        assert potential.shape[0] == self.rv_dim
        return np.clip(potential, 1e-12, None)

    def auto_coef(self, node_map, assign_policy=None):
        super().auto_coef(node_map, assign_policy)

        sum_i_alpha, unset_edge = self.__sum_neighbor_coef()
        if unset_edge:
            new_i_alpha = self.node_coef - \
                (1 - len(self.connections)) - sum_i_alpha
            self.connected_nodes[unset_edge].set_i_alpha(self.name, new_i_alpha)

    def __sum_neighbor_coef(self):
        sum_i_alpha = 0
        uninit_neighbor = None
        for node in self.connected_nodes.values():
            i_alpha = node.get_i_alpha(self.name)
            if i_alpha is not None:
                sum_i_alpha += i_alpha
            else:
                uninit_neighbor = node.name
        return sum_i_alpha, uninit_neighbor

    def cal_cnp_coef(self):
        self.coef_ready = True

        self.hat_c_i = self.node_coef
        for item in self.connections:
            self.hat_c_i += self.connected_nodes[item].node_coef

    def _make_message_first_term(self, recipient_node):
        recipient_index_in_var = self.search_msg_index(self.latest_message,
                                                       recipient_node.name)
        hat_c_ialpha = recipient_node.get_hat_c_ialpha(self.name)
        c_alpha = recipient_node.node_coef
        vals = [message.val for message in self.latest_message]

        with np.errstate(divide='raise'):
            if self.isconstrained:
                log_numerator = self.epsilon * np.log(self.constrained_marginal)
            else:
                potential_part = 1.0 / self.hat_c_i * np.log(self.potential)
                message_part = 1.0 / self.hat_c_i * \
                    np.log(np.clip(np.prod(vals, axis=0), 1e-12, None))
                log_numerator = potential_part + message_part
            clip_base = np.clip(vals[recipient_index_in_var], 1e-12, None)
            log_denominator = 1.0 / hat_c_ialpha * np.log(clip_base)

            log_base = c_alpha * (log_numerator - log_denominator)
            return np.exp(log_base)

    def make_message_bp(self, recipient_node):
        assert self.coef_ready,\
            f"{self.name} need to cal_cnp_coef by graph firstly"
        # first_term.shape equals (self.rv_dim,)
        first_term = self._make_message_first_term(recipient_node)
        assert first_term.shape[0] == self.rv_dim
        # second_term shape equals shape of recipient_node
        second_term = recipient_node.get_varnode_extra_term(self.name)
        assert second_term.shape == self.connected_nodes[recipient_node.name].potential.shape

        var_index_in_recipient = recipient_node.search_node_index(self.name)

        expanded_first_term = nd_expand(
            first_term, second_term.shape, var_index_in_recipient)

        return np.multiply(expanded_first_term, second_term)

    def make_message(self, recipient_node):
        return self.make_message_bp(recipient_node)

    def cal_bethe(self, margin):
        clip_margin = np.clip(margin, 1e-12, None)
        log_margin = np.log(clip_margin)
        entropy_term = -(self.node_degree - 1) * np.sum(margin * log_margin)
        clip_potential = np.clip(self.potential, 1e-12, None)
        potential_term = -np.sum(margin * np.log(clip_potential))
        return potential_term + entropy_term

    def marginal(self):
        if self.isconstrained:
            return self.constrained_marginal
        if self.message_inbox:
            vals = [message.val for message in self.latest_message]
            vals_prod = np.prod(vals, axis=0)
            prod = self.potential * vals_prod
            belief = np.power(prod, 1.0 / self.hat_c_i)
            return belief / np.sum(belief)

        return np.ones(self.rv_dim) / self.rv_dim

    def __eq__(self, value):
        if isinstance(value, type(self)):
            flag = []
            flag.append(self.isconstrained == value.isconstrained)
            flag.append(np.array_equal(self.constrained_marginal,
                                       value.constrained_marginal))
            if np.sum(flag) == len(flag):
                return super().__eq__(value)

        return False

    def plot(self, graph):
        if self.isconstrained:
            graph.add_node(self.name, color='red', style='filled')
        else:
            super().plot(graph)



from abc import abstractmethod

import numpy as np
from cbp.utils.message import Message

from .base_node import BaseNode


class DiscreteNode(BaseNode):
    def __init__(self, node_coef, potential):
        """[summary]

        :param node_coef: works for the norm-product algorithm
        :type node_coef: float
        :param potential: potential
        :type potential: ndarray or list
        """
        super().__init__()
        self.node_coef = node_coef
        self.coef_ready = False
        self.epsilon = 1
        self._potential = None
        self.potential = potential
        self.message_inbox = {}
        self.latest_message = []

    def auto_coef(self, node_map, assign_policy=None):
        if assign_policy is None:
            self.node_coef = 1.0 / len(node_map)
        else:
            self.node_coef = assign_policy(self, node_map)

        self.register_nodes(node_map)

    @property
    def potential(self):
        return self._potential

    @potential.setter
    def potential(self, potential):
        self._potential = self._check_potential(potential)

    @abstractmethod
    def _check_potential(self, potential) -> np.ndarray:
        """check potential before set node potential

        :param potential: input potential
        :type potential: np.ndarray
        :return: [description]
        :rtype: np.ndarray
        """

    def cal_cnp_coef(self):
        raise NotImplementedError(
            f"{self.__class__.__name__} is an abstract class")

    def make_init_message(self, recipient_node_name):
        if self.coef_ready:
            recipient_node = self.connected_nodes[recipient_node_name]
            message_dim = recipient_node.potential.shape
            return np.ones(message_dim)

        raise RuntimeError(
            f"Need to call cal_cnp_coef first for {self.name}")

    # keep all message looks urgly. convenient for debug and resource occupied
    # is not so huge
    def store_message(self, message):
        sender_name = message.sender.name
        self.message_inbox[sender_name] = message

        self.latest_message = list(self.message_inbox.values())

    def reset(self):
        self.message_inbox.clear()

    # TODO: FIXAPI NAME
    @abstractmethod
    def make_message(self, recipient_node) -> np.ndarray:
        """produce the val of message from current node to the recipient_node

        :param recipient_node: target node
        :type recipient_node: [type]
        :return: content of the message
        :rtype: np.ndarray
        """

    @abstractmethod
    def cal_bethe(self, margin) -> float:
        """calculate the bethe energy

        :return: bethe energy on this node
        :rtype: float
        """

    def send_message(self, recipient_node, verbose=False):
        """send message from this node to target node

        :param recipient_node: target node
        :param verbose: debug, defaults to False
        :type verbose: bool, optional
        """
        val = self.make_message(recipient_node)
        message = Message(self, val)
        recipient_node.store_message(message)
        if verbose:
            print(self.name + '->' + recipient_node.name)
            print(message.val)

    def sendin_message(self, verbose=False):
        for connected_node in self.connected_nodes.values():
            connected_node.send_message(self, verbose)

    def sendout_message(self, verbose=False):
        for connected_node in self.connected_nodes.values():
            self.send_message(connected_node, verbose)

    def search_msg_index(self, message_list, node_name):
        which_index = [i for i, message in enumerate(message_list)
                       if message.sender.name == node_name]
        if which_index:
            return which_index[0]

        raise RuntimeError(
            f"{node_name} do not appear in {self.name} message")

    def __eq__(self, value):
        flag = []
        flag.append((np.isclose(self.node_coef, value.node_coef)).all())
        flag.append(np.isclose(self.potential, value.potential).all())
        if np.sum(flag) == len(flag):
            return super().__eq__(value)

        return False



from dataclasses import dataclass
from .base_config import BaseConfig


@dataclass
class TestConfig(BaseConfig):
    @staticmethod
    def itsbp_schedule(cnt, leaf_nodes):
        if cnt == len(leaf_nodes) - 2:
            cnt = -1
            leaf_nodes.reverse()
        return cnt + 1
